@@ -77,6 +77,7 @@ struct a2dp_sep {
 };
 
 struct a2dp_setup_cb {
+	struct a2dp_setup *setup;
 	a2dp_select_cb_t select_cb;
 	a2dp_config_cb_t config_cb;
 	a2dp_stream_cb_t resume_cb;
@@ -92,7 +93,8 @@ struct a2dp_setup {
 	struct avdtp_remote_sep *rsep;
 	struct avdtp_stream *stream;
 	struct avdtp_error *err;
-	GSList *client_caps;
+	avdtp_set_configuration_cb setconf_cb;
+	GSList *caps;
 	gboolean reconfigure;
 	gboolean start;
 	GSList *cb;
@@ -123,27 +125,6 @@ static struct a2dp_setup *setup_ref(struct a2dp_setup *setup)
 	return setup;
 }
 
-static void setup_free(struct a2dp_setup *s)
-{
-	DBG("%p", s);
-	setups = g_slist_remove(setups, s);
-	if (s->session)
-		avdtp_unref(s->session);
-	g_slist_foreach(s->cb, (GFunc) g_free, NULL);
-	g_slist_free(s->cb);
-	g_free(s);
-}
-
-static void setup_unref(struct a2dp_setup *setup)
-{
-	setup->ref--;
-
-	DBG("%p: ref=%d", setup, setup->ref);
-
-	if (setup->ref <= 0)
-		setup_free(setup);
-}
-
 static struct audio_device *a2dp_get_dev(struct avdtp *session)
 {
 	bdaddr_t src, dst;
@@ -153,26 +134,90 @@ static struct audio_device *a2dp_get_dev(struct avdtp *session)
 	return manager_find_device(NULL, &src, &dst, NULL, FALSE);
 }
 
+static struct a2dp_setup *setup_new(struct avdtp *session)
+{
+	struct audio_device *dev;
+	struct a2dp_setup *setup;
+
+	dev = a2dp_get_dev(session);
+	if (!dev) {
+		error("Unable to create setup");
+		return NULL;
+	}
+
+	setup = g_new0(struct a2dp_setup, 1);
+	setup->session = avdtp_ref(session);
+	setup->dev = a2dp_get_dev(session);
+	setups = g_slist_append(setups, setup);
+
+	return setup;
+}
+
+static void setup_free(struct a2dp_setup *s)
+{
+	DBG("%p", s);
+
+	setups = g_slist_remove(setups, s);
+	if (s->session)
+		avdtp_unref(s->session);
+	g_slist_foreach(s->cb, (GFunc) g_free, NULL);
+	g_slist_free(s->cb);
+	g_slist_foreach(s->caps, (GFunc) g_free, NULL);
+	g_slist_free(s->caps);
+	g_free(s);
+}
+
+static void setup_unref(struct a2dp_setup *setup)
+{
+	setup->ref--;
+
+	DBG("%p: ref=%d", setup, setup->ref);
+
+	if (setup->ref > 0)
+		return;
+
+	setup_free(setup);
+}
+
+static struct a2dp_setup_cb *setup_cb_new(struct a2dp_setup *setup)
+{
+	struct a2dp_setup_cb *cb;
+
+	cb = g_new0(struct a2dp_setup_cb, 1);
+	cb->setup = setup;
+	cb->id = ++cb_id;
+
+	setup->cb = g_slist_append(setup->cb, cb);
+	return cb;
+}
+
+static void setup_cb_free(struct a2dp_setup_cb *cb)
+{
+	struct a2dp_setup *setup = cb->setup;
+
+	setup->cb = g_slist_remove(setup->cb, cb);
+	setup_unref(cb->setup);
+	g_free(cb);
+}
+
 static gboolean finalize_config(struct a2dp_setup *s)
 {
 	GSList *l;
 	struct avdtp_stream *stream = s->err ? NULL : s->stream;
 
-	setup_ref(s);
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
+		l = l->next;
+
 		if (!cb->config_cb)
 			continue;
 
 		cb->config_cb(s->session, s->sep, stream, s->err,
 							cb->user_data);
-		cb->config_cb = NULL;
-		setup_unref(s);
+		setup_cb_free(cb);
 	}
 
-	setup_unref(s);
-
 	return FALSE;
 }
 
@@ -180,7 +225,7 @@ static gboolean finalize_config_errno(struct a2dp_setup *s, int err)
 {
 	struct avdtp_error avdtp_err;
 
-	avdtp_error_init(&avdtp_err, AVDTP_ERROR_ERRNO, -err);
+	avdtp_error_init(&avdtp_err, AVDTP_ERRNO, -err);
 	s->err = err ? &avdtp_err : NULL;
 
 	return finalize_config(s);
@@ -190,19 +235,17 @@ static gboolean finalize_resume(struct a2dp_setup *s)
 {
 	GSList *l;
 
-	setup_ref(s);
-
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb && cb->resume_cb) {
-			cb->resume_cb(s->session, s->err, cb->user_data);
-			cb->resume_cb = NULL;
-			setup_unref(s);
-		}
-	}
+		l = l->next;
 
-	setup_unref(s);
+		if (!cb->resume_cb)
+			continue;
+
+		cb->resume_cb(s->session, s->err, cb->user_data);
+		setup_cb_free(cb);
+	}
 
 	return FALSE;
 }
@@ -211,18 +254,18 @@ static gboolean finalize_suspend(struct a2dp_setup *s)
 {
 	GSList *l;
 
-	setup_ref(s);
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->suspend_cb) {
-			cb->suspend_cb(s->session, s->err, cb->user_data);
-			cb->suspend_cb = NULL;
-			setup_unref(s);
-		}
+		l = l->next;
+
+		if (!cb->suspend_cb)
+			continue;
+
+		cb->suspend_cb(s->session, s->err, cb->user_data);
+		setup_cb_free(cb);
 	}
 
-	setup_unref(s);
 	return FALSE;
 }
 
@@ -230,28 +273,28 @@ static gboolean finalize_suspend_errno(struct a2dp_setup *s, int err)
 {
 	struct avdtp_error avdtp_err;
 
-	avdtp_error_init(&avdtp_err, AVDTP_ERROR_ERRNO, -err);
+	avdtp_error_init(&avdtp_err, AVDTP_ERRNO, -err);
 	s->err = err ? &avdtp_err : NULL;
 
 	return finalize_suspend(s);
 }
 
-static gboolean finalize_select(struct a2dp_setup *s, GSList *caps)
+static gboolean finalize_select(struct a2dp_setup *s)
 {
 	GSList *l;
 
-	setup_ref(s);
-	for (l = s->cb; l != NULL; l = l->next) {
+	for (l = s->cb; l != NULL; ) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->select_cb) {
-			cb->select_cb(s->session, s->sep, caps, cb->user_data);
-			cb->select_cb = NULL;
-			setup_unref(s);
-		}
+		l = l->next;
+
+		if (!cb->select_cb)
+			continue;
+
+		cb->select_cb(s->session, s->sep, s->caps, cb->user_data);
+		setup_cb_free(cb);
 	}
 
-	setup_unref(s);
 	return FALSE;
 }
 
@@ -269,6 +312,20 @@ static struct a2dp_setup *find_setup_by_session(struct avdtp *session)
 	return NULL;
 }
 
+static struct a2dp_setup *a2dp_setup_get(struct avdtp *session)
+{
+	struct a2dp_setup *setup;
+
+	setup = find_setup_by_session(session);
+	if (!setup) {
+		setup = setup_new(session);
+		if (!setup)
+			return NULL;
+	}
+
+	return setup_ref(setup);
+}
+
 static struct a2dp_setup *find_setup_by_dev(struct audio_device *dev)
 {
 	GSList *l;
@@ -311,26 +368,65 @@ static void stream_state_changed(struct avdtp_stream *stream,
 
 }
 
+static gboolean auto_config(gpointer data)
+{
+	struct a2dp_setup *setup = data;
+	struct avdtp_error *err = NULL;
+
+	/* Check if configuration was aborted */
+	if (setup->sep->stream == NULL)
+		return FALSE;
+
+	if (setup->err != NULL) {
+		err = setup->err;
+		goto done;
+	}
+
+	avdtp_stream_add_cb(setup->session, setup->stream,
+				stream_state_changed, setup->sep);
+
+	if (setup->sep->type == AVDTP_SEP_TYPE_SOURCE)
+		sink_new_stream(setup->dev, setup->session, setup->stream);
+	else
+		source_new_stream(setup->dev, setup->session, setup->stream);
+
+done:
+	if (setup->setconf_cb)
+		setup->setconf_cb(setup->session, setup->stream, setup->err);
+
+	finalize_config(setup);
+
+	if (err)
+		g_free(err);
+
+	setup_unref(setup);
+
+	return FALSE;
+}
+
 static gboolean sbc_setconf_ind(struct avdtp *session,
-				struct avdtp_local_sep *sep,
-				struct avdtp_stream *stream,
-				GSList *caps, uint8_t *err,
-				uint8_t *category, void *user_data)
+					struct avdtp_local_sep *sep,
+					struct avdtp_stream *stream,
+					GSList *caps,
+					avdtp_set_configuration_cb cb,
+					void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct audio_device *dev;
+	struct a2dp_setup *setup;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
 		DBG("Sink %p: Set_Configuration_Ind", sep);
 	else
 		DBG("Source %p: Set_Configuration_Ind", sep);
 
-	dev = a2dp_get_dev(session);
-	if (!dev) {
-		*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-		*category = 0x00;
+	setup = a2dp_setup_get(session);
+	if (!setup)
 		return FALSE;
-	}
+
+	a2dp_sep->stream = stream;
+	setup->sep = a2dp_sep;
+	setup->stream = stream;
+	setup->setconf_cb = cb;
 
 	/* Check valid settings */
 	for (; caps != NULL; caps = g_slist_next(caps)) {
@@ -340,9 +436,10 @@ static gboolean sbc_setconf_ind(struct avdtp *session,
 
 		if (cap->category == AVDTP_DELAY_REPORTING &&
 					!a2dp_sep->delay_reporting) {
-			*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-			*category = AVDTP_DELAY_REPORTING;
-			return FALSE;
+			setup->err = g_new(struct avdtp_error, 1);
+			avdtp_error_init(setup->err, AVDTP_DELAY_REPORTING,
+						AVDTP_UNSUPPORTED_CONFIGURATION);
+			goto done;
 		}
 
 		if (cap->category != AVDTP_MEDIA_CODEC)
@@ -360,18 +457,15 @@ static gboolean sbc_setconf_ind(struct avdtp *session,
 
 		if (sbc_cap->min_bitpool < MIN_BITPOOL ||
 					sbc_cap->max_bitpool > MAX_BITPOOL) {
-			*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-			*category = AVDTP_MEDIA_CODEC;
-			return FALSE;
+			setup->err = g_new(struct avdtp_error, 1);
+			avdtp_error_init(setup->err, AVDTP_MEDIA_CODEC,
+					AVDTP_UNSUPPORTED_CONFIGURATION);
+			goto done;
 		}
 	}
 
-	avdtp_stream_add_cb(session, stream, stream_state_changed, a2dp_sep);
-	a2dp_sep->stream = stream;
-
-	if (a2dp_sep->type == AVDTP_SEP_TYPE_SOURCE)
-		sink_new_stream(dev, session, stream);
-
+done:
+	g_idle_add(auto_config, setup);
 	return TRUE;
 }
 
@@ -439,43 +533,43 @@ static gboolean sbc_getcap_ind(struct avdtp *session, struct avdtp_local_sep *se
 }
 
 static gboolean mpeg_setconf_ind(struct avdtp *session,
-				struct avdtp_local_sep *sep,
-				struct avdtp_stream *stream,
-				GSList *caps, uint8_t *err,
-				uint8_t *category, void *user_data)
+					struct avdtp_local_sep *sep,
+					struct avdtp_stream *stream,
+					GSList *caps,
+					avdtp_set_configuration_cb cb,
+					void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct audio_device *dev;
+	struct a2dp_setup *setup;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
 		DBG("Sink %p: Set_Configuration_Ind", sep);
 	else
 		DBG("Source %p: Set_Configuration_Ind", sep);
 
-	dev = a2dp_get_dev(session);
-	if (!dev) {
-		*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-		*category = 0x00;
+	setup = a2dp_setup_get(session);
+	if (!setup)
 		return FALSE;
-	}
+
+	a2dp_sep->stream = stream;
+	setup->sep = a2dp_sep;
+	setup->stream = stream;
+	setup->setconf_cb = cb;
 
 	for (; caps != NULL; caps = g_slist_next(caps)) {
 		struct avdtp_service_capability *cap = caps->data;
 
 		if (cap->category == AVDTP_DELAY_REPORTING &&
 					!a2dp_sep->delay_reporting) {
-			*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-			*category = AVDTP_DELAY_REPORTING;
-			return FALSE;
+			setup->err = g_new(struct avdtp_error, 1);
+			avdtp_error_init(setup->err, AVDTP_DELAY_REPORTING,
+					AVDTP_UNSUPPORTED_CONFIGURATION);
+			goto done;
 		}
 	}
 
-	avdtp_stream_add_cb(session, stream, stream_state_changed, a2dp_sep);
-	a2dp_sep->stream = stream;
-
-	if (a2dp_sep->type == AVDTP_SEP_TYPE_SOURCE)
-		sink_new_stream(dev, session, stream);
-
+done:
+	g_idle_add(auto_config, setup);
 	return TRUE;
 }
 
@@ -536,26 +630,43 @@ static gboolean mpeg_getcap_ind(struct avdtp *session,
 	return TRUE;
 }
 
+static void endpoint_setconf_cb(struct media_endpoint *endpoint, void *ret,
+						int size, void *user_data)
+{
+	struct a2dp_setup *setup = user_data;
+
+	if (ret == NULL) {
+		setup->err = g_new(struct avdtp_error, 1);
+		avdtp_error_init(setup->err, AVDTP_MEDIA_CODEC,
+					AVDTP_UNSUPPORTED_CONFIGURATION);
+	}
+
+	auto_config(setup);
+}
+
 static gboolean endpoint_setconf_ind(struct avdtp *session,
-					struct avdtp_local_sep *sep,
-					struct avdtp_stream *stream,
-					GSList *caps, uint8_t *err,
-					uint8_t *category, void *user_data)
+						struct avdtp_local_sep *sep,
+						struct avdtp_stream *stream,
+						GSList *caps,
+						avdtp_set_configuration_cb cb,
+						void *user_data)
 {
 	struct a2dp_sep *a2dp_sep = user_data;
-	struct audio_device *dev;
+	struct a2dp_setup *setup;
 
 	if (a2dp_sep->type == AVDTP_SEP_TYPE_SINK)
 		DBG("Sink %p: Set_Configuration_Ind", sep);
 	else
 		DBG("Source %p: Set_Configuration_Ind", sep);
 
-	dev = a2dp_get_dev(session);
-	if (!dev) {
-		*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-		*category = 0x00;
+	setup = a2dp_setup_get(session);
+	if (!session)
 		return FALSE;
-	}
+
+	a2dp_sep->stream = stream;
+	setup->sep = a2dp_sep;
+	setup->stream = stream;
+	setup->setconf_cb = cb;
 
 	for (; caps != NULL; caps = g_slist_next(caps)) {
 		struct avdtp_service_capability *cap = caps->data;
@@ -564,9 +675,10 @@ static gboolean endpoint_setconf_ind(struct avdtp *session,
 
 		if (cap->category == AVDTP_DELAY_REPORTING &&
 					!a2dp_sep->delay_reporting) {
-			*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-			*category = AVDTP_DELAY_REPORTING;
-			return FALSE;
+			setup->err = g_new(struct avdtp_error, 1);
+			avdtp_error_init(setup->err, AVDTP_DELAY_REPORTING,
+					AVDTP_UNSUPPORTED_CONFIGURATION);
+			goto done;
 		}
 
 		if (cap->category != AVDTP_MEDIA_CODEC)
@@ -575,28 +687,26 @@ static gboolean endpoint_setconf_ind(struct avdtp *session,
 		codec = (struct avdtp_media_codec_capability *) cap->data;
 
 		if (codec->media_codec_type != a2dp_sep->codec) {
-			*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-			*category = AVDTP_MEDIA_CODEC;
-			return FALSE;
+			setup->err = g_new(struct avdtp_error, 1);
+			avdtp_error_init(setup->err, AVDTP_MEDIA_CODEC,
+					AVDTP_UNSUPPORTED_CONFIGURATION);
+			goto done;
 		}
 
-		ret = media_endpoint_set_configuration(a2dp_sep->endpoint, dev,
-					codec->data, cap->length - sizeof(*codec),
-					NULL, NULL);
+		ret = media_endpoint_set_configuration(a2dp_sep->endpoint,
+						setup->dev, codec->data,
+						cap->length - sizeof(*codec),
+						endpoint_setconf_cb, setup);
 		if (ret)
-			break;
+			return TRUE;
 
-		*err = AVDTP_UNSUPPORTED_CONFIGURATION;
-		*category = AVDTP_MEDIA_CODEC;
-		return FALSE;
+		avdtp_error_init(setup->err, AVDTP_MEDIA_CODEC,
+					AVDTP_UNSUPPORTED_CONFIGURATION);
+		break;
 	}
 
-	avdtp_stream_add_cb(session, stream, stream_state_changed, a2dp_sep);
-	a2dp_sep->stream = stream;
-
-	if (a2dp_sep->type == AVDTP_SEP_TYPE_SOURCE)
-		sink_new_stream(dev, session, stream);
-
+done:
+	g_idle_add(auto_config, setup);
 	return TRUE;
 }
 
@@ -647,6 +757,27 @@ static gboolean endpoint_getcap_ind(struct avdtp *session,
 	return TRUE;
 }
 
+static void endpoint_open_cb(struct media_endpoint *endpoint, void *ret,
+						int size, void *user_data)
+{
+	struct a2dp_setup *setup = user_data;
+	int err;
+
+	if (ret == NULL) {
+		setup->stream = NULL;
+		finalize_config_errno(setup, -EPERM);
+		return;
+	}
+
+	err = avdtp_open(setup->session, setup->stream);
+	if (err == 0)
+		return;
+
+	error("Error on avdtp_open %s (%d)", strerror(-err), -err);
+	setup->stream = NULL;
+	finalize_config_errno(setup, err);
+}
+
 static void setconf_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 				struct avdtp_stream *stream,
 				struct avdtp_error *err, void *user_data)
@@ -695,11 +826,14 @@ static void setconf_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 
 		if (media_endpoint_set_configuration(a2dp_sep->endpoint, dev,
 						codec->data, service->length -
-						sizeof(*codec), NULL, NULL) ==
-						FALSE) {
-			setup->stream = NULL;
-			finalize_config_errno(setup, -EPERM);
-		}
+						sizeof(*codec),
+						endpoint_open_cb, setup) ==
+						TRUE)
+			return;
+
+		setup->stream = NULL;
+		finalize_config_errno(setup, -EPERM);
+		return;
 	}
 
 	ret = avdtp_open(session, stream);
@@ -897,7 +1031,7 @@ static void suspend_cfm(struct avdtp *session, struct avdtp_local_sep *sep,
 	} else if (avdtp_start(session, a2dp_sep->stream) < 0) {
 		struct avdtp_error start_err;
 		error("avdtp_start failed");
-		avdtp_error_init(&start_err, AVDTP_ERROR_ERRNO, EIO);
+		avdtp_error_init(&start_err, AVDTP_ERRNO, EIO);
 		setup->err = err;
 		finalize_suspend(setup);
 	}
@@ -923,9 +1057,12 @@ static gboolean a2dp_reconfigure(gpointer data)
 	struct a2dp_sep *sep = setup->sep;
 	int posix_err;
 
+	if (!setup->rsep)
+		setup->rsep = avdtp_find_remote_sep(setup->session, sep->lsep);
+
 	posix_err = avdtp_set_configuration(setup->session, setup->rsep,
 						sep->lsep,
-						setup->client_caps,
+						setup->caps,
 						&setup->stream);
 	if (posix_err < 0) {
 		error("avdtp_set_configuration: %s", strerror(-posix_err));
@@ -1226,7 +1363,8 @@ int a2dp_register(DBusConnection *conn, const bdaddr_t *src, GKeyFile *config)
 {
 	int sbc_srcs = 1, sbc_sinks = 1;
 	int mpeg12_srcs = 0, mpeg12_sinks = 0;
-	gboolean source = TRUE, sink = FALSE, delay_reporting;
+	gboolean source = TRUE, sink = FALSE, socket = TRUE;
+	gboolean delay_reporting = FALSE;
 	char *str;
 	GError *err = NULL;
 	int i;
@@ -1258,9 +1396,20 @@ int a2dp_register(DBusConnection *conn, const bdaddr_t *src, GKeyFile *config)
 			source = FALSE;
 		if (strstr(str, "Source"))
 			sink = FALSE;
+		if (strstr(str, "Socket"))
+			socket = FALSE;
 		g_free(str);
 	}
 
+	/* Don't register any local sep if Socket is disabled */
+	if (socket == FALSE) {
+		sbc_srcs = 0;
+		sbc_sinks = 0;
+		mpeg12_srcs = 0;
+		mpeg12_sinks = 0;
+		goto proceed;
+	}
+
 	str = g_key_file_get_string(config, "A2DP", "SBCSources", &err);
 	if (err) {
 		DBG("audio.conf: %s", err->message);
@@ -1319,8 +1468,10 @@ proceed:
 		servers = g_slist_append(servers, server);
 	}
 
-	delay_reporting = g_key_file_get_boolean(config, "A2DP",
+	if (config)
+		delay_reporting = g_key_file_get_boolean(config, "A2DP",
 						"DelayReporting", NULL);
+
 	if (delay_reporting)
 		server->version = 0x0103;
 	else
@@ -1662,7 +1813,6 @@ static void select_cb(struct media_endpoint *endpoint, void *ret, int size,
 	struct a2dp_setup *setup = user_data;
 	struct avdtp_service_capability *media_transport, *media_codec;
 	struct avdtp_media_codec_capability *cap;
-	GSList *caps = NULL;
 
 	if (size < 0) {
 		DBG("Endpoint replied an invalid configuration");
@@ -1672,7 +1822,7 @@ static void select_cb(struct media_endpoint *endpoint, void *ret, int size,
 	media_transport = avdtp_service_cap_new(AVDTP_MEDIA_TRANSPORT,
 						NULL, 0);
 
-	caps = g_slist_append(caps, media_transport);
+	setup->caps = g_slist_append(setup->caps, media_transport);
 
 	cap = g_malloc0(sizeof(*cap) + size);
 	cap->media_type = AVDTP_MEDIA_TYPE_AUDIO;
@@ -1682,17 +1832,18 @@ static void select_cb(struct media_endpoint *endpoint, void *ret, int size,
 	media_codec = avdtp_service_cap_new(AVDTP_MEDIA_CODEC, cap,
 						sizeof(*cap) + size);
 
-	caps = g_slist_append(caps, media_codec);
+	setup->caps = g_slist_append(setup->caps, media_codec);
+	g_free(cap);
 
 done:
-	finalize_select(setup, caps);
+	finalize_select(setup);
 }
 
 static gboolean auto_select(gpointer data)
 {
 	struct a2dp_setup *setup = data;
 
-	finalize_select(setup, setup->client_caps);
+	finalize_select(setup);
 
 	return FALSE;
 }
@@ -1764,21 +1915,14 @@ unsigned int a2dp_select_capabilities(struct avdtp *session,
 		return 0;
 	}
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->select_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->rsep = avdtp_find_remote_sep(session, sep->lsep);
 
@@ -1791,7 +1935,7 @@ unsigned int a2dp_select_capabilities(struct avdtp *session,
 	endpoint in the configuration file */
 	if (sep->endpoint == NULL) {
 		if (!select_capabilities(session, setup->rsep,
-					&setup->client_caps)) {
+					&setup->caps)) {
 			error("Unable to auto select remote SEP capabilities");
 			goto fail;
 		}
@@ -1811,8 +1955,7 @@ unsigned int a2dp_select_capabilities(struct avdtp *session,
 		return cb_data->id;
 
 fail:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 
 }
@@ -1855,24 +1998,23 @@ unsigned int a2dp_config(struct avdtp *session, struct a2dp_sep *sep,
 
 	DBG("a2dp_config: selected SEP %p", sep->lsep);
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->config_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->stream = sep->stream;
-	setup->client_caps = caps;
+
+	/* Copy given caps if they are different than current caps */
+	if (setup->caps != caps) {
+		g_slist_foreach(setup->caps, (GFunc) g_free, NULL);
+		g_slist_free(setup->caps);
+		setup->caps = g_slist_copy(caps);
+	}
 
 	switch (avdtp_sep_get_state(sep->lsep)) {
 	case AVDTP_STATE_IDLE:
@@ -1937,8 +2079,7 @@ unsigned int a2dp_config(struct avdtp *session, struct a2dp_sep *sep,
 	return cb_data->id;
 
 failed:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 }
 
@@ -1948,21 +2089,14 @@ unsigned int a2dp_resume(struct avdtp *session, struct a2dp_sep *sep,
 	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->resume_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->stream = sep->stream;
 
@@ -1996,8 +2130,7 @@ unsigned int a2dp_resume(struct avdtp *session, struct a2dp_sep *sep,
 	return cb_data->id;
 
 failed:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 }
 
@@ -2007,21 +2140,14 @@ unsigned int a2dp_suspend(struct avdtp *session, struct a2dp_sep *sep,
 	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
 
-	cb_data = g_new0(struct a2dp_setup_cb, 1);
+	setup = a2dp_setup_get(session);
+	if (!setup)
+		return 0;
+
+	cb_data = setup_cb_new(setup);
 	cb_data->suspend_cb = cb;
 	cb_data->user_data = user_data;
-	cb_data->id = ++cb_id;
 
-	setup = find_setup_by_session(session);
-	if (!setup) {
-		setup = g_new0(struct a2dp_setup, 1);
-		setup->session = avdtp_ref(session);
-		setup->dev = a2dp_get_dev(session);
-		setups = g_slist_append(setups, setup);
-	}
-
-	setup_ref(setup);
-	setup->cb = g_slist_append(setup->cb, cb_data);
 	setup->sep = sep;
 	setup->stream = sep->stream;
 
@@ -2048,44 +2174,39 @@ unsigned int a2dp_suspend(struct avdtp *session, struct a2dp_sep *sep,
 	return cb_data->id;
 
 failed:
-	setup_unref(setup);
-	cb_id--;
+	setup_cb_free(cb_data);
 	return 0;
 }
 
 gboolean a2dp_cancel(struct audio_device *dev, unsigned int id)
 {
-	struct a2dp_setup_cb *cb_data;
 	struct a2dp_setup *setup;
 	GSList *l;
 
-	DBG("a2dp_cancel()");
-
 	setup = find_setup_by_dev(dev);
 	if (!setup)
 		return FALSE;
 
-	for (cb_data = NULL, l = setup->cb; l != NULL; l = g_slist_next(l)) {
+	for (l = setup->cb; l != NULL; l = g_slist_next(l)) {
 		struct a2dp_setup_cb *cb = l->data;
 
-		if (cb->id == id) {
-			cb_data = cb;
-			break;
+		if (cb->id != id)
+			continue;
+
+		setup_ref(setup);
+		setup_cb_free(cb);
+
+		if (!setup->cb) {
+			DBG("aborting setup %p", setup);
+			avdtp_abort(setup->session, setup->stream);
+			return TRUE;
 		}
+
+		setup_unref(setup);
+		return TRUE;
 	}
 
-	if (!cb_data)
-		error("a2dp_cancel: no matching callback with id %u", id);
-
-	setup->cb = g_slist_remove(setup->cb, cb_data);
-	g_free(cb_data);
-
-	if (setup->cb)
-		return TRUE;
-
-	avdtp_abort(setup->session, setup->stream);
-
-	return TRUE;
+	return FALSE;
 }
 
 gboolean a2dp_sep_lock(struct a2dp_sep *sep, struct avdtp *session)
