@@ -45,9 +45,16 @@
 #include <bluetooth/sdp_lib.h>
 
 #include "textfile.h"
+#include "adapter.h"
+#include "device.h"
 #include "glib-helper.h"
 #include "storage.h"
 
+struct match {
+	GSList *keys;
+	char *pattern;
+};
+
 static inline int create_filename(char *buf, size_t size,
 				const bdaddr_t *bdaddr, const char *name)
 {
@@ -1241,28 +1248,91 @@ int write_blocked(const bdaddr_t *local, const bdaddr_t *remote,
 int write_device_services(const bdaddr_t *sba, const bdaddr_t *dba,
 							const char *services)
 {
-       char filename[PATH_MAX + 1], addr[18];
+	char filename[PATH_MAX + 1], addr[18];
 
-       create_filename(filename, PATH_MAX, sba, "primary");
+	create_filename(filename, PATH_MAX, sba, "primary");
 
-       create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 
-       ba2str(dba, addr);
+	ba2str(dba, addr);
 
-       return textfile_put(filename, addr, services);
+	return textfile_put(filename, addr, services);
+}
+
+static void filter_keys(char *key, char *value, void *data)
+{
+	struct match *match = data;
+	const char *address = match->pattern;
+
+	/* Each key contains: MAC#handle*/
+	if (strncasecmp(key, address, 17) == 0)
+		match->keys = g_slist_append(match->keys, g_strdup(key));
+}
+
+int delete_device_service(const bdaddr_t *sba, const bdaddr_t *dba)
+{
+	GSList *l;
+	struct match match;
+	char filename[PATH_MAX + 1], address[18];
+	int err;
+
+	create_filename(filename, PATH_MAX, sba, "primary");
+
+	memset(address, 0, sizeof(address));
+	ba2str(dba, address);
+
+	err = textfile_del(filename, address);
+	if (err < 0)
+		return err;
+
+	/* Deleting all characteristics of a given address */
+	memset(&match, 0, sizeof(match));
+	match.pattern = address;
+
+	create_filename(filename, PATH_MAX, sba, "characteristic");
+	err = textfile_foreach(filename, filter_keys, &match);
+	if (err < 0)
+		return err;
+
+	for (l = match.keys; l; l = l->next) {
+		const char *key = l->data;
+		textfile_del(filename, key);
+	}
+
+	g_slist_foreach(match.keys, (GFunc) g_free, NULL);
+	g_slist_free(match.keys);
+
+	/* Deleting all attributes values of a given address */
+	memset(&match, 0, sizeof(match));
+	match.pattern = address;
+
+	create_filename(filename, PATH_MAX, sba, "attributes");
+	err = textfile_foreach(filename, filter_keys, &match);
+	if (err < 0)
+		return err;
+
+	for (l = match.keys; l; l = l->next) {
+		const char *key = l->data;
+		textfile_del(filename, key);
+	}
+
+	g_slist_foreach(match.keys, (GFunc) g_free, NULL);
+	g_slist_free(match.keys);
+
+	return 0;
 }
 
 char *read_device_services(const bdaddr_t *sba, const bdaddr_t *dba)
 {
-       char filename[PATH_MAX + 1], addr[18];
+	char filename[PATH_MAX + 1], addr[18];
 
-       create_filename(filename, PATH_MAX, sba, "primary");
+	create_filename(filename, PATH_MAX, sba, "primary");
 
-       create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
 
-       ba2str(dba, addr);
+	ba2str(dba, addr);
 
-       return textfile_caseget(filename, addr);
+	return textfile_caseget(filename, addr);
 }
 
 int write_device_characteristics(const bdaddr_t *sba, const bdaddr_t *dba,
@@ -1296,3 +1366,68 @@ char *read_device_characteristics(const bdaddr_t *sba, const bdaddr_t *dba,
 
 	return textfile_caseget(filename, key);
 }
+
+int write_device_attribute(const bdaddr_t *sba, const bdaddr_t *dba,
+					uint16_t handle, const char *chars)
+{
+	char filename[PATH_MAX + 1], addr[18], key[23];
+
+	create_filename(filename, PATH_MAX, sba, "attributes");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	ba2str(dba, addr);
+
+	snprintf(key, sizeof(key), "%17s#%04X", addr, handle);
+
+	return textfile_put(filename, key, chars);
+}
+
+int read_device_attributes(const bdaddr_t *sba, textfile_cb func, void *data)
+{
+	char filename[PATH_MAX + 1];
+
+	create_filename(filename, PATH_MAX, sba, "attributes");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	return textfile_foreach(filename, func, data);
+}
+
+int write_device_type(const bdaddr_t *sba, const bdaddr_t *dba,
+						device_type_t type)
+{
+	char filename[PATH_MAX + 1], addr[18], chars[3];
+
+	create_filename(filename, PATH_MAX, sba, "types");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	ba2str(dba, addr);
+
+	snprintf(chars, sizeof(chars), "%2.2X", type);
+
+	return textfile_put(filename, addr, chars);
+}
+
+device_type_t read_device_type(const bdaddr_t *sba, const bdaddr_t *dba)
+{
+	char filename[PATH_MAX + 1], addr[18], *chars;
+	device_type_t type;
+
+	create_filename(filename, PATH_MAX, sba, "types");
+
+	create_file(filename, S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH);
+
+	ba2str(dba, addr);
+
+	chars = textfile_caseget(filename, addr);
+	if (chars == NULL)
+		return DEVICE_TYPE_UNKNOWN;
+
+	type = strtol(chars, NULL, 16);
+
+	free(chars);
+
+	return type;
+}
